# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

# pyright: strict

from ._implementations import Zen, ZenStore, default_to_config, store, zen

__all__ = ["zen", "Zen", "store", "ZenStore", "default_to_config"]



# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT
import warnings
from collections import UserList
from dataclasses import fields, is_dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from hydra import initialize
from hydra._internal.callbacks import Callbacks
from hydra.core.config_store import ConfigStore
from hydra.core.global_hydra import GlobalHydra
from hydra.core.utils import JobReturn, run_job
from hydra.plugins.sweeper import Sweeper
from hydra.types import HydraContext, RunMode
from omegaconf import DictConfig, ListConfig, OmegaConf
from typing_extensions import Literal, TypeAlias

from hydra_zen._compatibility import SUPPORTS_VERSION_BASE
from hydra_zen._hydra_overloads import instantiate
from hydra_zen.typing._implementations import DataClass_, InstOrType

T = TypeVar("T", bound=Any)
HydraPrimitives: TypeAlias = Union[None, int, float, bool, str, Dict[str, str]]

if TYPE_CHECKING:  # pragma: no cover
    # branching needed to deal with pyright type-completeness complaints
    TUserList: TypeAlias = UserList[Any]
else:
    TUserList = UserList


class _NotSet:  # pragma: no cover
    pass


T1 = TypeVar("T1", bound=HydraPrimitives)


class hydra_list(TUserList, Generic[T1]):
    """Signals that a sequence is provided as a single configured value (i.e. it is not
    to be iterated over during a multirun)"""

    pass


T2 = TypeVar("T2", bound=Union[HydraPrimitives, hydra_list[HydraPrimitives]])


class multirun(TUserList, Generic[T2]):
    """Signals that a sequence is to be iterated over in a multirun"""

    pass


def _safe_name(x: Any) -> str:
    return getattr(x, "__name__", str(x))


def value_check(
    name: str,
    value: T,
    type_: Union[type, Tuple[type, ...]],
) -> T:
    """
    For internal use only.

    Used to check the type of `value`. Numerical types can also be bound-checked.

    Examples
    --------
    >>> value_check("x", 1, type_=str)
    TypeError: `x` must be of type(s) `str`, got 1 (type: int)

    Raises
    ------
    TypeError"""
    # check internal params
    assert isinstance(name, str), name

    if not isinstance(value, type_):
        raise TypeError(
            f"`{name}` must be of type(s) "
            f"`{_safe_name(type_)}`, got {value} (type: {_safe_name(type(value))})"
        )

    return cast(T, value)


OverrideValues: TypeAlias = Union[
    HydraPrimitives,
    multirun[Union[HydraPrimitives, hydra_list[HydraPrimitives]]],
    hydra_list[HydraPrimitives],
]
OverrideDict: TypeAlias = Mapping[str, OverrideValues]


def _process_dict_overrides(overrides: OverrideDict) -> List[str]:
    """Convert dict overrides to a list of Hydra CLI compatible args"""
    launch_overrides = []
    for k, v in overrides.items():
        if v is None:
            v = "null"

        value_check(
            k,
            v,
            type_=(int, float, bool, str, dict, multirun, hydra_list),
        )
        if isinstance(v, multirun):
            v = ",".join(str(item) for item in v)

        launch_overrides.append(f"{k}={v}")
    return launch_overrides


def _store_config(
    cfg: Union[DataClass_, Type[DataClass_], DictConfig, ListConfig, Mapping[Any, Any]],
    config_name: str = "hydra_launch",
) -> str:
    """Stores configuration object in Hydra's ConfigStore.

    Parameters
    ----------
    cfg : Union[DataClass_, DictConfig, Mapping]
        A configuration as a dataclass, configuration object, or a dictionary.

    config_name : str (default: hydra_launch)
        The configuration name used to store the configuration.

    Returns
    -------
    config_name : str
        The configuration name used to store the default configuration.

    Notes
    -----
    The input configuration is registered in the Hydra ConfigStore [1]_ using a
    user-provided config name.

    References
    ----------
    .. [1] https://hydra.cc/docs/tutorials/structured_config/config_store
    """
    cs = ConfigStore().instance()
    cs.store(name=config_name, node=cfg)
    return config_name


@overload
def launch(
    config: Union[InstOrType[DataClass_], Mapping[str, Any]],
    task_function: Callable[[Any], Any],
    overrides: Optional[Union[OverrideDict, List[str]]] = ...,
    multirun: Literal[False] = ...,
    version_base: Optional[Union[str, Type[_NotSet]]] = ...,
    to_dictconfig: bool = ...,
    config_name: str = ...,
    job_name: str = ...,
    with_log_configuration: bool = ...,
    **override_kwargs: OverrideValues,
) -> JobReturn:
    ...


@overload
def launch(
    config: Union[InstOrType[DataClass_], Mapping[str, Any]],
    task_function: Callable[[Any], Any],
    overrides: Optional[Union[OverrideDict, List[str]]] = ...,
    multirun: Literal[True] = ...,
    version_base: Optional[Union[str, Type[_NotSet]]] = ...,
    to_dictconfig: bool = ...,
    config_name: str = ...,
    job_name: str = ...,
    with_log_configuration: bool = ...,
    **override_kwargs: OverrideValues,
) -> Any:
    ...


def launch(
    config: Union[InstOrType[DataClass_], Mapping[str, Any]],
    task_function: Callable[[Any], Any],
    overrides: Optional[Union[OverrideDict, List[str]]] = None,
    multirun: bool = False,
    version_base: Optional[Union[str, Type[_NotSet]]] = _NotSet,
    to_dictconfig: bool = False,
    config_name: str = "zen_launch",
    job_name: str = "zen_launch",
    with_log_configuration: bool = True,
    **override_kwargs: OverrideValues,
) -> Union[JobReturn, Any]:
    r"""
    Launches a Hydra job from a Python function rather than a CLI.

    `launch` is designed to closely match the interface of the standard Hydra CLI.
    For example, launching a Hydra job from the CLI via::

       $ python my_task.py job/group=group_name job.group.param=1

    corresponds to the following usage of `launch`:

       >>> job = launch(config, task_function, overrides=["job/group=group_name", "job.group.param=1"])

    Parameters
    ----------
    config : DataClass_ | Type[DataClass_] | Mapping[str, Any]
        A config that will be passed to ``task_function``.

    task_function : Callable[[DictConfig], Any]
        The function that Hydra will execute. Its input will be ``config``, which
        has been modified via the specified ``overrides``

    overrides : Optional[Union[OverrideMapping, List[str]]] (default: None)
        If provided, sets/overrides values in ``config``. See [1]_ and [2]_
        for a detailed discussion of the "grammar" supported by ``overrides``.

    multirun : bool (default: False)
        Launch a Hydra multi-run ([3]_).

    version_base : Optional[str], optional (default=not-specified)
        Available starting with Hydra 1.2.0.
        - If the `version_base parameter` is not specified, Hydra 1.x will use defaults compatible with version 1.1. Also in this case, a warning is issued to indicate an explicit version_base is preferred.
        - If the `version_base parameter` is `None`, then the defaults are chosen for the current minor Hydra version. For example for Hydra 1.2, then would imply `config_path=None` and `hydra.job.chdir=False`.
        - If the `version_base` parameter is an explicit version string like "1.1", then the defaults appropriate to that version are used.

    to_dictconfig : bool (default: False)
        If ``True``, convert a ``dataclasses.dataclass`` to a ``omegaconf.DictConfig``.
        Note, this will remove Hydra's cabability for validation with structured
        configurations.

    config_name : str (default: "zen_launch")
        Name of the stored configuration in Hydra's ConfigStore API.

    job_name : str (default: "zen_launch")

    with_log_configuration : bool (default: True)
        If ``True``, enables the configuration of the logging subsystem from the loaded
        config.

    **override_kwargs : OverrideValues
        Keyword arguments to override existing configuration values.  Note, this only
        works when the configuration value name is a valid Python identifier; e.g.,
        this does not support adding (`+param`) values.

    Returns
    -------
    result : hydra.core.utils.JobReturn | Any
        If ``multirun is False``:
            A ``JobReturn`` object storing the results of the Hydra experiment via the
            following attributes
                - ``cfg``: Reflects ``config``
                - ``overrides``: Reflects ``overrides``
                - ``return_value``: The return value of the task function
                - ``hydra_cfg``: The Hydra configuration object
                - ``working_dir``: The experiment working directory
                - ``task_name``: The task name of the Hydra job
                - ``status``: A ``JobStatus`` enum reporting whether or not the job completed successfully
        Else:
            Return values of all launched jobs (depends on the Sweeper implementation).

    References
    ----------
    .. [1] https://hydra.cc/docs/advanced/override_grammar/basic
    .. [2] https://hydra.cc/docs/configure_hydra/intro
    .. [3] https://hydra.cc/docs/tutorials/basic/running_your_app/multi-run

    Examples
    --------

    **Basic usage**

    Let's define and launch a trivial Hydra app.

    >>> from hydra_zen import make_config, launch, to_yaml

    First, we will define a config, which determines the configurable interface to our
    "app". For the purpose of example, we'll design the "interface" of this config to accept
    two configurable parameters: ``a`` and ``b``.

    >>> Conf = make_config("a", "b")

    Our task function accepts the config as an input and uses it to run some generic functionality.
    For simplicity's sake, let's design this task function to: convert the job's config to a
    yaml-formatted string, print it, and then return the string.

    >>> def task_fn(cfg):
    ...     out = to_yaml(cfg)  # task's input config, converted to yaml-string
    ...     print(out)
    ...     return out

    Now, let's use `launch` to run this task function via Hydra, using particular configured
    values (or, "overrides") for ``a`` and ``b``.

    >>> job_out = launch(Conf, task_fn, a=1, b='foo')
    a: 1
    b: foo

    Let's inspect ``job_out`` to see the ways that it summarizes the results of this job.

    >>> job_out.return_value  # the value returned by `task_fn`
    'a: 1\nb: foo\n'

    >>> job_out.working_dir  # where the job's outputs, logs, and configs are saved
    'outputs/2021-10-19/15-27-11'

    >>> job_out.cfg  # the particular config used to run our task-function
    {'a': 1, 'b': 'foo'}

    >>> job_out.overrides  # the overrides that we provides
    ['a=1', "b='foo'"]

    >>> job_out.status  # the job's completion status
    <JobStatus.COMPLETED: 1>

    **Launching a multirun job**

    We can launch multiple runs of our task-function, using various configured values.
    Let's launch a multirun that sweeps over three configurations

    >>> (outputs,) = launch(
    ...     Conf,
    ...     task_fn,
    ...     a="1,2,3",
    ...     b="bar",
    ...     multirun=True,
    ... )
    [2021-10-19 17:50:07,334][HYDRA] Launching 3 jobs locally
    [2021-10-19 17:50:07,334][HYDRA] 	#0 : a=1 b='bar'
    a: 1
    b: bar
    [2021-10-19 17:50:07,434][HYDRA] 	#1 : a=2 b='bar'
    a: 2
    b: bar
    [2021-10-19 17:50:07,535][HYDRA] 	#2 : a=3 b='bar'
    a: 3
    b: bar

    ``outputs`` contains three corresponding ``JobReturns`` instances.

    >>> len(outputs)
    3
    >>> [j.cfg for j in outputs]
    [{'a': 1, 'b': 'bar'}, {'a': 2, 'b': 'bar'}, {'a': 3, 'b': 'bar'}]

    Each run's outputs, logs, and configs are saved to separate working directories

    >>> [j.working_dir for j in outputs]
    ['multirun/2021-10-19/17-50-07\\0',
    'multirun/2021-10-19/17-50-07\\1',
    'multirun/2021-10-19/17-50-07\\2']

    **Launching with quoted overrides**

    Some of the Hydra CLI override syntax cannot be specified as keyword arguments. In such cases we can instead provide a list or a dict with quoted overrides.

    >>> job_out = launch(Conf, task_fn, a=1, b="foo", overrides={"+c": 22})
    a: 1
    b: foo
    c: 22
    >>> job_out.overrides  # the overrides that we provides
    ['a=1', 'b=foo', '+c=22']

    >>> launch(Conf, task_fn, overrides=["a=1", "b='foo'", "+c=22"])
    a: 1
    b: foo
    c: 22

    **Using hydra_zen.multirun**

    Multi-run values can be specified directly, without having to form a quoted multi-run string, by using the `hydra_zen.multi_run` list to store the values.

    >>> import random
    >>> from hydra_zen import launch, instantiate, make_config, multirun
    >>> values_for_experiment = [random.uniform(0, 1) for i in range(10)]

    >>> jobs = launch(
    ...     make_config(),
    ...     instantiate,
    ...     overrides={
    ...          "+param": multirun(values_for_experiment)
    ...     },
    ...     multirun=True
    ... )

    If, instead, you want to configure a list as a single value - not to be iterated
    over in a multirun - you can instead use `hydra_zen.hydra_list`.
    """

    # used for check below
    _num_dataclass_fields = 0
    if is_dataclass(config):
        _num_dataclass_fields = len(fields(config))

    # store config in ConfigStore
    if to_dictconfig and is_dataclass(config):
        # convert Dataclass to a DictConfig
        dictconfig = OmegaConf.create(
            OmegaConf.to_container(OmegaConf.structured(config))
        )
        config_name = _store_config(dictconfig, config_name)
    else:
        config_name = _store_config(config, config_name)

    # allow user to provide a dictionary of override values
    # instead of just a list of strings
    overrides = overrides if overrides is not None else []
    if isinstance(overrides, Mapping):
        overrides = _process_dict_overrides(overrides)

    override_kwargs_list = _process_dict_overrides(override_kwargs)
    overrides += override_kwargs_list

    # Initializes Hydra and add the config_path to the config search path
    with initialize(
        config_path=None,
        job_name=job_name,
        **(
            {}
            if (not SUPPORTS_VERSION_BASE or version_base is _NotSet)
            else {"version_base": version_base}
        ),
    ):
        # taken from hydra.compose with support for MULTIRUN
        gh = GlobalHydra.instance()
        assert gh.hydra is not None

        # Load configuration
        cfg = gh.hydra.compose_config(
            config_name=config_name,
            overrides=overrides,
            run_mode=RunMode.RUN if not multirun else RunMode.MULTIRUN,
            from_shell=False,
            with_log_configuration=with_log_configuration,
        )

        callbacks = Callbacks(cfg)
        run_start = (
            callbacks.on_run_start if not multirun else callbacks.on_multirun_start
        )
        run_start(config=cfg, config_name=config_name)

        hydra_context = HydraContext(
            config_loader=gh.config_loader(), callbacks=callbacks
        )

        if not multirun:
            job = run_job(
                hydra_context=hydra_context,
                task_function=task_function,
                config=cfg,
                job_dir_key="hydra.run.dir",
                job_subdir_key=None,
                configure_logging=with_log_configuration,
            )
            callbacks.on_run_end(config=cfg, config_name=config_name, job_return=job)

            # access the result to trigger an exception in case the job failed.
            _ = job.return_value
        else:
            # Instantiate sweeper without using Hydra's Plugin discovery (Zen!)
            sweeper = instantiate(cfg.hydra.sweeper)
            assert isinstance(sweeper, Sweeper)
            sweeper.setup(
                config=cfg,
                hydra_context=hydra_context,
                task_function=task_function,
            )

            task_overrides = OmegaConf.to_container(
                cfg.hydra.overrides.task, resolve=False
            )
            assert isinstance(task_overrides, list)
            job = sweeper.sweep(arguments=task_overrides)
            callbacks.on_multirun_end(config=cfg, config_name=config_name)

    if is_dataclass(config):
        _num_dataclass_fields_after = len(fields(config))
        if (
            _num_dataclass_fields_after == 0
            and _num_dataclass_fields_after < _num_dataclass_fields
        ):
            warnings.warn(
                "Your dataclass-based config was mutated by this run. If you just "
                "executed with a `hydra/launcher` that utilizes cloudpickle (e.g., "
                "hydra-submitit-launcher), there is a known issue with dataclasses "
                "(see: https://github.com/cloudpipe/cloudpickle/issues/386). You will "
                "have to restart your interactive environment to run `launch` again. "
                "To avoid this issue you can use the `launch` option: "
                "`to_dictconfig=True`."
            )

    return job



# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

from ._implementations import (
    Builds,
    DataclassOptions,
    HydraPartialBuilds,
    Just,
    Partial,
    PartialBuilds,
    StoreEntry,
    SupportedPrimitive,
    ZenConvert,
    ZenPartialBuilds,
    ZenWrappers,
)

__all__ = [
    "Builds",
    "DataclassOptions",
    "Just",
    "Partial",
    "PartialBuilds",
    "SupportedPrimitive",
    "ZenWrappers",
    "ZenPartialBuilds",
    "HydraPartialBuilds",
    "ZenConvert",
    "StoreEntry",
]



# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT
import functools
from collections import Counter, deque
from dataclasses import InitVar, dataclass, field
from functools import partial
from pathlib import Path, PosixPath, WindowsPath
from typing import Any, Tuple, Type, TypeVar, cast

from hydra_zen._compatibility import ZEN_SUPPORTED_PRIMITIVES
from hydra_zen.typing import Builds, Partial, PartialBuilds

from ._implementations import ZEN_VALUE_CONVERSION, builds, sanitized_default_value
from ._utils import get_obj_path

_T = TypeVar("_T")


@dataclass(unsafe_hash=True)
class ConfigComplex:
    real: Any
    imag: Any
    _target_: str = field(default=get_obj_path(complex), init=False)


def convert_complex(value: complex) -> Builds[Type[complex]]:
    return cast(Builds[Type[complex]], ConfigComplex(real=value.real, imag=value.imag))


ZEN_VALUE_CONVERSION[complex] = convert_complex


@dataclass(unsafe_hash=True)
class ConfigPath:
    _args_: Tuple[str]
    _target_: str = field(default=get_obj_path(Path), init=False)


if Path in ZEN_SUPPORTED_PRIMITIVES:  # pragma: no cover

    def convert_path(value: Path) -> Builds[Type[Path]]:
        return cast(Builds[Type[Path]], ConfigPath(_args_=(str(value),)))  # type: ignore

    ZEN_VALUE_CONVERSION[Path] = convert_path
    ZEN_VALUE_CONVERSION[PosixPath] = convert_path
    ZEN_VALUE_CONVERSION[WindowsPath] = convert_path


def _unpack_partial(value: Partial[_T]) -> PartialBuilds[Type[_T]]:
    target = cast(Type[_T], value.func)
    return builds(target, *value.args, **value.keywords, zen_partial=True)()


@dataclass(unsafe_hash=True)
class ConfigFromTuple:
    _args_: Tuple[Any, ...]
    _target_: str

    def __post_init__(self):
        self._args_ = (
            sanitized_default_value(
                tuple(self._args_),
                convert_dataclass=True,
                allow_zen_conversion=True,
                structured_conf_permitted=True,
            ),
        )


@dataclass(unsafe_hash=True)
class ConfigFromDict:
    _args_: Any
    _target_: str

    def __post_init__(self):
        self._args_ = (
            sanitized_default_value(
                dict(self._args_),
                convert_dataclass=True,
                allow_zen_conversion=True,
                structured_conf_permitted=True,
            ),
        )


@dataclass(unsafe_hash=True)
class ConfigRange:
    start: InitVar[int]
    stop: InitVar[int]
    step: InitVar[int]
    _target_: str = field(default=get_obj_path(range), init=False)
    _args_: Tuple[int, ...] = field(default=(), init=False, repr=False)

    def __post_init__(self, start, stop, step):
        self._args_ = (start, stop, step)


ZEN_VALUE_CONVERSION[set] = partial(ConfigFromTuple, _target_=get_obj_path(set))
ZEN_VALUE_CONVERSION[frozenset] = partial(
    ConfigFromTuple, _target_=get_obj_path(frozenset)
)
ZEN_VALUE_CONVERSION[deque] = partial(ConfigFromTuple, _target_=get_obj_path(deque))

if bytes in ZEN_SUPPORTED_PRIMITIVES:  # pragma: no cover
    ZEN_VALUE_CONVERSION[bytes] = partial(ConfigFromTuple, _target_=get_obj_path(bytes))

ZEN_VALUE_CONVERSION[bytearray] = partial(
    ConfigFromTuple, _target_=get_obj_path(bytearray)
)
ZEN_VALUE_CONVERSION[range] = lambda value: ConfigRange(
    value.start, value.stop, value.step
)
ZEN_VALUE_CONVERSION[Counter] = partial(ConfigFromDict, _target_=get_obj_path(Counter))
ZEN_VALUE_CONVERSION[functools.partial] = _unpack_partial



# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT
from typing import TYPE_CHECKING

from ._hydra_overloads import (
    MISSING,
    instantiate,
    load_from_yaml,
    save_as_yaml,
    to_yaml,
)
from ._launch import hydra_list, launch, multirun
from .structured_configs import (
    ZenField,
    builds,
    hydrated_dataclass,
    just,
    make_config,
    make_custom_builds_fn,
    mutable_value,
)
from .structured_configs._implementations import get_target
from .structured_configs._type_guards import is_partial_builds, uses_zen_processing
from .wrapper import ZenStore, store, zen

__all__ = [
    "builds",
    "hydrated_dataclass",
    "just",
    "mutable_value",
    "get_target",
    "MISSING",
    "instantiate",
    "load_from_yaml",
    "save_as_yaml",
    "to_yaml",
    "make_config",
    "ZenField",
    "make_custom_builds_fn",
    "launch",
    "is_partial_builds",
    "uses_zen_processing",
    "zen",
    "hydra_list",
    "multirun",
    "store",
    "ZenStore",
]

if not TYPE_CHECKING:
    try:
        from ._version import version as __version__
    except ImportError:
        __version__ = "unknown version"
else:  # pragma: no cover
    __version__: str



# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT


class HydraZenException(Exception):
    """Generic parent class for exceptions thrown by hydra-zen."""


class HydraZenDeprecationWarning(HydraZenException, FutureWarning):
    """A deprecation warning issued by hydra-zen.

    Notes
    -----
    This is a subclass of FutureWarning, rather than DeprecationWarning, so
    that the warnings that it emits are not filtered by default.
    """


class HydraZenValidationError(HydraZenException):
    pass


class HydraZenUnsupportedPrimitiveError(HydraZenValidationError, ValueError):
    pass



# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT
# pyright: strict
from typing import FrozenSet

from typing_extensions import Final

from hydra_zen._compatibility import HYDRA_SUPPORTS_PARTIAL
from hydra_zen.funcs import get_obj, zen_processing
from hydra_zen.structured_configs import _utils

# Hydra-specific fields
TARGET_FIELD_NAME: Final[str] = "_target_"
PARTIAL_FIELD_NAME: Final[str] = "_partial_"
RECURSIVE_FIELD_NAME: Final[str] = "_recursive_"
CONVERT_FIELD_NAME: Final[str] = "_convert_"
POS_ARG_FIELD_NAME: Final[str] = "_args_"
DEFAULTS_LIST_FIELD_NAME: Final[str] = "defaults"

_names = [
    TARGET_FIELD_NAME,
    RECURSIVE_FIELD_NAME,
    CONVERT_FIELD_NAME,
    POS_ARG_FIELD_NAME,
]

if HYDRA_SUPPORTS_PARTIAL:  # pragma: no cover
    _names.append(PARTIAL_FIELD_NAME)

HYDRA_FIELD_NAMES: FrozenSet[str] = frozenset(_names)

del _names

# hydra-zen-specific fields
ZEN_PROCESSING_LOCATION: Final[str] = _utils.get_obj_path(zen_processing)
GET_OBJ_LOCATION: Final[str] = _utils.get_obj_path(get_obj)
ZEN_TARGET_FIELD_NAME: Final[str] = "_zen_target"
ZEN_PARTIAL_FIELD_NAME: Final[str] = "_zen_partial"
META_FIELD_NAME: Final[str] = "_zen_exclude"
ZEN_WRAPPERS_FIELD_NAME: Final[str] = "_zen_wrappers"
JUST_FIELD_NAME: Final[str] = "path"



# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT
from collections import Counter, deque
from enum import Enum
from functools import partial
from pathlib import Path, PosixPath, WindowsPath
from typing import NamedTuple, Set

import hydra
import omegaconf
from typing_extensions import Final

NoneType = type(None)


class Version(NamedTuple):
    major: int
    minor: int
    patch: int


def _get_version(ver_str: str) -> Version:
    # Not for general use. Tested only for Hydra and OmegaConf
    # version string styles

    splits = ver_str.split(".")[:3]
    if not len(splits) == 3:  # pragma: no cover
        raise ValueError(f"Version string {ver_str} couldn't be parsed")

    major, minor = (int(v) for v in splits[:2])
    patch_str, *_ = splits[-1].split("rc")

    return Version(major=major, minor=minor, patch=int(patch_str))


OMEGACONF_VERSION: Final = _get_version(omegaconf.__version__)
HYDRA_VERSION: Final = _get_version(hydra.__version__)

SUPPORTS_VERSION_BASE = HYDRA_VERSION >= (1, 2, 0)

# OmegaConf issue 830 describes a bug associated with structured configs
# composed via inheritance, where the child's attribute is a default-factory
# and the parent's corresponding attribute is not.
# We provide downstream workarounds until an upstream fix is released.
#
# Uncomment dynamic setting once OmegaConf merges fix:
# https://github.com/omry/omegaconf/pull/832
PATCH_OMEGACONF_830: Final = OMEGACONF_VERSION < Version(2, 2, 1)

# Hydra's instantiate API now supports partial-instantiation, indicated
# by a `_partial_ = True` attribute.
# https://github.com/facebookresearch/hydra/pull/1905
HYDRA_SUPPORTS_PARTIAL: Final = Version(1, 1, 1) < HYDRA_VERSION

HYDRA_SUPPORTS_NESTED_CONTAINER_TYPES: Final = OMEGACONF_VERSION >= Version(2, 2, 0)
HYDRA_SUPPORTS_BYTES: Final = OMEGACONF_VERSION >= Version(2, 2, 0)
HYDRA_SUPPORTS_Path: Final = OMEGACONF_VERSION >= Version(2, 2, 1)

# Indicates primitive types permitted in type-hints of structured configs
HYDRA_SUPPORTED_PRIMITIVE_TYPES: Final = {int, float, bool, str, Enum}
# Indicates types of primitive values permitted in configs
HYDRA_SUPPORTED_PRIMITIVES = {int, float, bool, str, list, tuple, dict, NoneType}
ZEN_SUPPORTED_PRIMITIVES: Set[type] = {
    set,
    frozenset,
    complex,
    partial,
    bytearray,
    deque,
    Counter,
    range,
}

HYDRA_SUPPORTS_LIST_INSTANTIATION = HYDRA_VERSION >= Version(1, 1, 2)


if HYDRA_SUPPORTS_BYTES:  # pragma: no cover
    HYDRA_SUPPORTED_PRIMITIVES.add(bytes)
    HYDRA_SUPPORTED_PRIMITIVE_TYPES.add(bytes)
else:  # pragma: no cover
    ZEN_SUPPORTED_PRIMITIVES.add(bytes)

_path_types = {Path, PosixPath, WindowsPath}

if HYDRA_SUPPORTS_Path:  # pragma: no cover
    HYDRA_SUPPORTED_PRIMITIVES.update(_path_types)
    HYDRA_SUPPORTED_PRIMITIVE_TYPES.add(Path)
else:  # pragma: no cover
    ZEN_SUPPORTED_PRIMITIVES.update(_path_types)

del _path_types

HYDRA_SUPPORTS_OBJECT_CONVERT = HYDRA_VERSION >= Version(1, 3, 0)



# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT
# pyright: strict
from dataclasses import MISSING
from functools import partial
from typing import TYPE_CHECKING, Any, Type, Union

from typing_extensions import TypeGuard

from hydra_zen._compatibility import HYDRA_SUPPORTS_PARTIAL
from hydra_zen.funcs import get_obj, zen_processing
from hydra_zen.structured_configs._utils import safe_name
from hydra_zen.typing import Builds, Just, PartialBuilds
from hydra_zen.typing._implementations import DataClass_, HasTarget

from ._globals import (
    JUST_FIELD_NAME,
    PARTIAL_FIELD_NAME,
    TARGET_FIELD_NAME,
    ZEN_PARTIAL_FIELD_NAME,
    ZEN_PROCESSING_LOCATION,
    ZEN_TARGET_FIELD_NAME,
)

__all__ = ["is_partial_builds", "uses_zen_processing", "is_dataclass"]

# We need to check if things are Builds, Just, PartialBuilds to a higher
# fidelity than is provided by `isinstance(..., <Protocol>)`. I.e. we want to
# check that the desired attributes *and* that their values match those of the
# protocols. Failing to heed this would, for example, lead to any `Builds` that
# happens to have a `path` attribute to be treated as `Just` in `get_target`.
#
# The following functions perform these desired checks. Note that they do not
# require that the provided object be a dataclass; this enables compatibility
# with omegaconf containers.
#
# These are not part of the public API for now, but they may be in the future.


def safe_getattr(obj: Any, field: str, *default: Any) -> Any:
    # We must access slotted class-attributes from a dataclass type
    # via its `__dataclass_fields__`. Otherwise we will get a member
    # descriptor

    assert len(default) < 2
    if (
        hasattr(obj, "__slots__")
        and isinstance(obj, type)
        and is_dataclass(obj)
        and field in obj.__slots__  # type: ignore
    ):
        try:
            _field = obj.__dataclass_fields__[field]
            if _field.default_factory is not MISSING or _field.default is MISSING:
                raise AttributeError

            return _field.default

        except (KeyError, AttributeError):
            if default:
                return default[0]

            raise AttributeError(
                f"type object '{safe_name(obj)}' has no attribute '{field}'"
            )

    return getattr(obj, field, *default)


def _get_target(x: HasTarget) -> Any:
    return safe_getattr(x, TARGET_FIELD_NAME)


def is_builds(x: Any) -> TypeGuard[Builds[Any]]:
    return hasattr(x, TARGET_FIELD_NAME)


def is_just(x: Any) -> TypeGuard[Just[Any]]:
    if is_builds(x) and hasattr(x, JUST_FIELD_NAME):
        attr = _get_target(x)
        if attr == _get_target(Just) or attr is get_obj:
            return True
        else:
            # ensures we convert this branch in tests
            return False
    return False


if TYPE_CHECKING:  # pragma: no cover

    def is_dataclass(obj: Any) -> TypeGuard[Union[DataClass_, Type[DataClass_]]]:
        ...

else:
    from dataclasses import is_dataclass


def is_old_partial_builds(x: Any) -> bool:  # pragma: no cover
    # We don't care about coverage here.
    # This will only be used in `get_target` and we'll be sure to cover that branch
    if is_builds(x) and hasattr(x, "_partial_target_"):
        attr = _get_target(x)
        if (attr == "hydra_zen.funcs.partial" or attr is partial) and is_just(
            safe_getattr(x, "_partial_target_")
        ):
            return True
        else:
            # ensures we cover this branch in tests
            return False
    return False


def uses_zen_processing(x: Any) -> TypeGuard[Builds[Any]]:
    """Returns `True` if the input is a targeted structured config that relies on
    zen-processing features during its instantiation process. See notes for more details

    Parameters
    ----------
    x : Any

    Returns
    -------
    uses_zen : bool

    Notes
    -----
    In order to support zen :ref:`meta-fields <meta-field>` and
    :ref:`zen wrappers <zen-wrapper>`, hydra-zen redirects Hydra to an intermediary
    function – `hydra_zen.funcs.zen_processing` – during instantiation; i.e.
    `zen_processing` is made to be the `_target_` of the config and `_zen_target`
    indicates the object that is ultimately being configured for instantiation.

    Examples
    --------
    >>> from hydra_zen import builds, uses_zen_processing, to_yaml
    >>> ConfA = builds(dict, a=1)
    >>> ConfB = builds(dict, a=1, zen_partial=True)
    >>> ConfC = builds(dict, a=1, zen_wrappers=lambda x: x)
    >>> ConfD = builds(dict, a=1, zen_meta=dict(hidden_field=None))
    >>> ConfE = builds(dict, a=1, zen_meta=dict(hidden_field=None), zen_partial=True)
    >>> uses_zen_processing(ConfA)
    False
    >>> uses_zen_processing(ConfB)
    False
    >>> uses_zen_processing(ConfC)
    True
    >>> uses_zen_processing(ConfD)
    True
    >>> uses_zen_processing(ConfE)
    True

    Demonstrating the indirection that is used to facilitate zen-processing features.

    >>> print(to_yaml(ConfE))
    _target_: hydra_zen.funcs.zen_processing
    _zen_target: builtins.dict
    _zen_partial: true
    _zen_exclude:
    - hidden_field
    a: 1
    hidden_field: null
    """
    if not is_builds(x) or not hasattr(x, ZEN_TARGET_FIELD_NAME):
        return False

    attr = _get_target(x)
    if attr != ZEN_PROCESSING_LOCATION and attr is not zen_processing:
        return False
    return True


def is_partial_builds(x: Any) -> TypeGuard[PartialBuilds[Any]]:
    """
    Returns `True` if the input is a targeted structured config that entails partial
    instantiation, either via `_partial_=True` [1]_ or via `_zen_partial=True`.

    Parameters
    ----------
    x : Any

    Returns
    -------
    is_partial_config : bool

    References
    ----------
    .. [1] https://hydra.cc/docs/advanced/instantiate_objects/overview/#partial-instantiation

    See Also
    --------
    uses_zen_processing

    Examples
    --------
    >>> from hydra_zen import is_partial_builds

    An example involving a basic structured config

    >>> from dataclasses import dataclass
    >>> @dataclass
    ... class A:
    ...     _target_ : str = 'builtins.int'
    ...     _partial_ : bool = True
    >>> is_partial_builds(A)
    True
    >>> is_partial_builds(A(_partial_=False))
    False

    An example of a config that leverages partial instantiation via zen-processing

    >>> from hydra_zen import builds, uses_zen_processing, instantiate
    >>> Conf = builds(int, 0, zen_partial=True, zen_meta=dict(a=1))
    >>> hasattr(Conf, "_partial_")
    False
    >>> uses_zen_processing(Conf)
    True
    >>> is_partial_builds(Conf)
    True
    >>> instantiate(Conf)
    functools.partial(<class 'int'>, 0)
    """
    if is_builds(x):
        return (
            # check if partial'd config via Hydra
            HYDRA_SUPPORTS_PARTIAL
            and safe_getattr(x, PARTIAL_FIELD_NAME, False) is True
        ) or (
            # check if partial'd config via `zen_processing`
            uses_zen_processing(x)
            and (safe_getattr(x, ZEN_PARTIAL_FIELD_NAME, False) is True)
        )
    return False



# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

from ._implementations import builds, hydrated_dataclass, mutable_value
from ._just import just
from ._make_config import ZenField, make_config
from ._make_custom_builds import make_custom_builds_fn

__all__ = [
    "builds",
    "just",
    "hydrated_dataclass",
    "mutable_value",
    "make_config",
    "ZenField",
    "make_custom_builds_fn",
]



# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT
from typing import Any, Callable, FrozenSet, Optional, Type, TypeVar, Union, overload

from typing_extensions import Literal

from hydra_zen.structured_configs import _utils
from hydra_zen.typing import Builds, DataclassOptions, Just
from hydra_zen.typing._implementations import _HydraPrimitive  # type: ignore
from hydra_zen.typing._implementations import _SupportedViaBuilds  # type: ignore
from hydra_zen.typing._implementations import AllConvert, DataClass_, ZenConvert

from ._implementations import sanitized_default_value
from ._utils import merge_settings
from ._value_conversion import ConfigComplex

# pyright: strict
T = TypeVar("T")
TD = TypeVar("TD", bound=DataClass_)
TC = TypeVar("TC", bound=Callable[..., Any])
TP = TypeVar("TP", bound=_HydraPrimitive)
TB = TypeVar("TB", bound=Union[_SupportedViaBuilds, FrozenSet[Any]])

__all__ = ["just"]


_JUST_CONVERT_SETTINGS = AllConvert(dataclass=True)


@overload
def just(
    obj: TP,
    *,
    zen_convert: Optional[ZenConvert] = ...,
    hydra_recursive: Optional[bool] = ...,
    hydra_convert: Optional[Literal["none", "partial", "all", "object"]] = ...,
    zen_dataclass: Optional[DataclassOptions] = ...,
) -> TP:
    ...


@overload
def just(
    obj: complex,
    *,
    zen_convert: Optional[ZenConvert] = ...,
    hydra_recursive: Optional[bool] = ...,
    hydra_convert: Optional[Literal["none", "partial", "all", "object"]] = ...,
    zen_dataclass: Optional[DataclassOptions] = ...,
) -> ConfigComplex:
    ...


@overload
def just(
    obj: TC,
    *,
    zen_convert: Optional[ZenConvert] = ...,
    hydra_recursive: Optional[bool] = ...,
    hydra_convert: Optional[Literal["none", "partial", "all", "object"]] = ...,
    zen_dataclass: Optional[DataclassOptions] = ...,
) -> Just[TC]:
    ...


@overload
def just(
    obj: TB,
    *,
    zen_convert: Optional[ZenConvert] = ...,
    hydra_recursive: Optional[bool] = ...,
    hydra_convert: Optional[Literal["none", "partial", "all", "object"]] = ...,
    zen_dataclass: Optional[DataclassOptions] = ...,
) -> Builds[Type[TB]]:
    ...


@overload
def just(
    obj: TD,
    *,
    zen_convert: Literal[None] = ...,
    hydra_recursive: Optional[bool] = ...,
    hydra_convert: Optional[Literal["none", "partial", "all", "object"]] = ...,
    zen_dataclass: Optional[DataclassOptions] = ...,
) -> Type[Builds[Type[TD]]]:
    ...


@overload
def just(
    obj: DataClass_,
    *,
    zen_convert: ZenConvert,
    hydra_recursive: Optional[bool] = ...,
    hydra_convert: Optional[Literal["none", "partial", "all", "object"]] = ...,
    zen_dataclass: Optional[DataclassOptions] = ...,
) -> Any:
    ...


@overload
def just(
    obj: Any,
    *,
    zen_convert: Optional[ZenConvert] = ...,
    hydra_recursive: Optional[bool] = ...,
    hydra_convert: Optional[Literal["none", "partial", "all", "object"]] = ...,
    zen_dataclass: Optional[DataclassOptions] = ...,
) -> Any:
    ...


def just(
    obj: Any,
    *,
    zen_convert: Optional[ZenConvert] = None,
    hydra_recursive: Optional[bool] = None,
    hydra_convert: Optional[Literal["none", "partial", "all", "object"]] = None,
    zen_dataclass: Optional[DataclassOptions] = None,
) -> Any:
    """`just(obj)` returns a config that, when instantiated, just returns `obj`.

    I.e., `instantiate(just(obj)) == obj`

    `just` is designed to be idempotent. I.e., `just(obj) == just(just(obj))`

    Parameters
    ----------
    obj : Callable[..., Any] | HydraSupportedPrimitive | ZenSupportedPrimitive
        A type (e.g. a class-object), function-object, or a value that is either
        supported by Hydra or has auto-config support via hydra-zen.

    zen_convert : Optional[ZenConvert]
        A dictionary that modifies hydra-zen's value and type conversion behavior.
        Consists of the following optional key-value pairs (:ref:`zen-convert`):

        - `dataclass` : `bool` (default=True):
            If `True` any dataclass type/instance without a
            `_target_` field is automatically converted to a targeted config
            that will instantiate to that type/instance. Otherwise the dataclass
            type/instance will be passed through as-is.


    hydra_recursive : Optional[bool], optional (default=True)
        If ``True``, then Hydra will recursively instantiate all other
        hydra-config objects nested within this config [2]_.

        If ``None``, the ``_recursive_`` attribute is not set on the resulting config.

    hydra_convert : Optional[Literal["none", "partial", "all", "object"]], optional (default="none")
        Determines how Hydra treats the non-primitive, omegaconf-specific objects
        during instantiateion [3]_.

        - ``"none"``: No conversion occurs; omegaconf containers are passed through (Default)
        - ``"partial"``: ``DictConfig`` and ``ListConfig`` objects converted to ``dict`` and
          ``list``, respectively. Structured configs and their fields are passed without conversion.
        - ``"all"``: All passed objects are converted to dicts, lists, and primitives, without
          a trace of OmegaConf containers.
        - ``"object"``: Passed objects are converted to dict and list. Structured Configs are converted to instances of the backing dataclass / attr class.

        If ``None``, the ``_convert_`` attribute is not set on the resulting config.

    zen_dataclass : Optional[DataclassOptions]
        A dictionary that can specify any option that is supported by
        :py:func:`dataclasses.make_dataclass` other than `fields`.
        The default value for `unsafe_hash` is `True`.

        These options are only relevant when the input to `just` is a dataclass
        instance. Otherwise, `just` does not utilize these options when auto-generating
        configs.

        Additionally, the `module` field can be specified to enable pickle
        compatibility. See `hydra_zen.typing.DataclassOptions` for details.

    Returns
    -------
    out : HydraSupportedPrimitive | Builds[Type[obj]]
        ``out`` is ``obj`` unchanged if ``obj`` is supported natively by Hydra, otherwise ``out`` is a dynamically-generated dataclass type or instance.

    Raises
    ------
    hydra_zen.errors.HydraZenUnsupportedPrimitiveError

    References
    ----------
    .. [1] https://hydra.cc/docs/tutorials/structured_config/intro/
    .. [2] https://hydra.cc/docs/advanced/instantiate_objects/overview/#recursive-instantiation
    .. [3] https://hydra.cc/docs/advanced/instantiate_objects/overview/#parameter-conversion-strategies

    See Also
    --------
    builds : Create a targeted structured config designed to "build" a particular object.
    make_config: Creates a general config with customized field names, default values, and annotations.

    Notes
    -----
    Here, a "config" is a dynamically-generated dataclass type that is designed to be
    compatible with Hydra [1]_.

    The configs produced by ``just(<type_or_func>)`` introduce an explicit dependency
    on hydra-zen. I.e. hydra-zen must be installed in order to instantiate any config
    that used `just`.

    Examples
    --------
    **Basic usage**

    >>> from hydra_zen import just, instantiate, to_yaml

    `just`, called on a value of a type that is natively supported by Hydra,
    will return that value unchanged:

    >>> just(1)
    1
    >>> just({"a": False})
    {"a": False}

    `just` can be used to create a config that will simply import a class-object or function when the config is instantiated by Hydra.


    >>> class A: pass
    >>> def my_func(x: int): pass

    >>> instantiate(just(A)) is A
    True
    >>> instantiate(just(my_func)) is my_func
    True

    `just` dynamically generates dataclass types, a.k.a structured configs, to describe
    ``obj``

    >>> just(A)()
    Just_A(_target_='hydra_zen.funcs.get_obj', path='__main__.A')
    >>> just(my_func)()
    Just_my_func(_target_='hydra_zen.funcs.get_obj', path='__main__.my_func')

    **Auto-config support**

    Calling `just` on a value of a type that has
    :ref:`special support from hydra-zen <additional-types>`
    will return a structured config instance that, when instantiated by Hydra, returns
    that value.

    >>> just(1+2j)
    ConfigComplex(real=1.0, imag=2.0, _target_='builtins.complex')
    >>> instantiate(just(1+2j))
    (1+2j)

    >>> just({1, 2, 3})
    Builds_set(_target_='builtins.set', _args_=((1, 2, 3),))
    >>> instantiate(just({1, 2, 3}))
    {1, 2, 3}

    By default, `just` will convert a dataclass instance to a corresponding targeted
    config. This behavior can be modified via :ref:`zen-convert settings<zen-convert>`.

    >>> from dataclasses import dataclass
    >>> @dataclass
    ... class B:
    ...     x: complex
    >>>
    >>> instantiate(just(B)) is B
    True
    >>> instantiate(just(B(2+3j))) == B(2+3j)
    True

    `just` operates recursively within sequences, mappings, and dataclass fields.

    >>> z_dict = {'a': [3-4j, 1+2j]}
    >>> assert instantiate(just(z_dict)) == z_dict
    >>>
    >>> from typing import Any
    >>> @dataclass
    ... class C:
    ...     x: Any
    >>>
    >>> instantiate(just(C(B(2+3j))))
    C(B(2+3j))

    **Auto-Application of just**

    Both `builds` and `make_config` will automatically (and recursively) apply
    `just` to all configured values. E.g. in the following example `just` will be
    applied to both  the complex-valued the list and to ``sum``.

    >>> from hydra_zen import make_config
    >>> Conf2 = make_config(data=[1+2j, 2+3j], reduction_fn=sum)

    >>> print(to_yaml(Conf2))
    data:
    - real: 1.0
      imag: 2.0
      _target_: builtins.complex
    - real: 2.0
      imag: 3.0
      _target_: builtins.complex
    reduction_fn:
      _target_: hydra_zen.funcs.get_obj
      path: builtins.sum

    >>> conf = instantiate(Conf2)
    >>> conf.reduction_fn(conf.data)
    (3+5j)
    """
    convert_settings = merge_settings(zen_convert, _JUST_CONVERT_SETTINGS)
    del zen_convert
    _utils.validate_hydra_options(
        hydra_recursive=hydra_recursive, hydra_convert=hydra_convert
    )
    if zen_dataclass is None:
        zen_dataclass = {}

    return sanitized_default_value(
        obj,
        allow_zen_conversion=True,
        structured_conf_permitted=True,
        field_name="",
        error_prefix="",
        convert_dataclass=convert_settings["dataclass"],
        hydra_convert=hydra_convert,
        hydra_recursive=hydra_recursive,
        zen_dataclass=_utils.parse_dataclass_options(zen_dataclass),
    )
